#include "FSMDeterminizer.h"

namespace determinization {

set<State> FSMDeterminizer::findToStates(const set<State>& nfsmStates, const Symbol& input) {
    const set<TransitionFSM>& nfsmTransitions = this->nondeterministicFSM->getTransitions();
    set<State> targetStates;
    for (set<State>::iterator state = nfsmStates.begin(); state != nfsmStates.end(); state++) {
        for (set<TransitionFSM>::iterator transition = nfsmTransitions.begin();
                transition != nfsmTransitions.end();
                transition++) {
            if (transition->getFrom() == *state && transition->getInput() == input) {
                targetStates.insert(transition->getTo());
            }
        }
    }
    return targetStates;
}


FSMDeterminizer::FSMDeterminizer(FSM* automaton) {
    this->nondeterministicFSM = automaton;
}


FSM* FSMDeterminizer::determinize() {
    this->deterministicFSM = new FSM();
    DeterminizationUtils::copyInputAlphabet(*this->nondeterministicFSM, *this->deterministicFSM);
    map<string, DeterminizationUtils::StateData> states;
    
    const set<State>& nfsmInitialStates = this->nondeterministicFSM->getInitialStates();
    const State& initialState = DeterminizationUtils::getOrCreateState(nfsmInitialStates, states, *this->deterministicFSM);
    
    while (true) {
        bool allStatesMarked = true;
        for (map<string, DeterminizationUtils::StateData>::iterator stateIter = states.begin();
                stateIter != states.end();
                stateIter++) {
            allStatesMarked &= stateIter->second.isMarked;
        }
        if (allStatesMarked) {
            break;
        }
        
        DeterminizationUtils::StateData* unmarkedStateData;
        for (map<string, DeterminizationUtils::StateData>::iterator stateIter = states.begin();
                stateIter != states.end();
                stateIter++) {
            DeterminizationUtils::StateData& stateData = stateIter->second;
            if (!stateData.isMarked) {
                unmarkedStateData = &stateData;
                break;
            }
        }
        
        const set<Symbol>& nfsmInputAlphabet = this->nondeterministicFSM->getInputAlphabet();
        for (set<Symbol>::iterator input = nfsmInputAlphabet.begin(); input != nfsmInputAlphabet.end(); input++) {
            const set<State>& nfsmToStates = this->findToStates(unmarkedStateData->originalStates, *input);
            if (nfsmToStates.size() > 0) {
                const State& targetState = DeterminizationUtils::getOrCreateState(nfsmToStates, states, *this->deterministicFSM);
                TransitionFSM transition(unmarkedStateData->state, *input, targetState);
                this->deterministicFSM->addTransition(transition);
            }
        }
        
        unmarkedStateData->isMarked = true;
    }
    
    this->deterministicFSM->addInitialState(initialState);
    
    for (map<string, DeterminizationUtils::StateData>::iterator stateIter = states.begin();
            stateIter != states.end();
            stateIter++) {
        const DeterminizationUtils::StateData& stateData = stateIter->second;
        if (DeterminizationUtils::containSomeFinalStateOfAutomaton(stateData.originalStates, *this->nondeterministicFSM)) {
            this->deterministicFSM->addFinalState(stateData.state);
        }
    }
    
    return this->deterministicFSM;
}

} /* namespace determinization */
