#include "VPADeterminizer.h"

namespace determinization {
    
const Symbol VPADeterminizer::BOTTOM_OF_STACK_SYMBOL("⊥");
    

string VPADeterminizer::buildStateName(const SComponent& s, const RComponent& r) {
    string name = "{";
    for (set<StatesPair>::const_iterator statePair = s.statesPairs.begin(); statePair != s.statesPairs.end(); statePair++) {
        if (statePair != s.statesPairs.begin()) {
            name += ", ";
        }
        name += "('" + statePair->state1.getName() + "', '" + statePair->state2.getName() + "')";
    }
    name += "}, {";
    for (set<State>::const_iterator state = r.states.begin(); state != r.states.end(); state++) {
        if (state != r.states.begin()) {
            name += ", ";
        }
        name += "'" + state->getName() + "'";
    }
    name += "}";
    return name;
}


const State& VPADeterminizer::getOrCreateState(const SComponent& s, const RComponent& r) {
    string stateName = this->buildStateName(s, r);
    
    map<string, StateData>::iterator statesIter = this->states.find(stateName);
    if (statesIter != this->states.end()) {
        return statesIter->second.state;
    }
    
    State state(stateName);
    StateData stateData(state, s, r);
    StateData& insertedData = this->states.insert(pair<string, StateData>(stateName, stateData)).first->second;
    this->deterministicVPA->addState(state);
    return insertedData.state;
}


const VPADeterminizer::SComponent VPADeterminizer::getSComponentWithStatesIdentity() {
    set<State> states = this->nondeterministicVPA->getStates();
    set<StatesPair> statesPairs;
    for (set<State>::const_iterator state = states.begin(); state != states.end(); state++) {
        statesPairs.insert(StatesPair(*state, *state));
    }
    return SComponent(statesPairs);
}


void VPADeterminizer::divideInputAlphabet() {
    this->internalSymbols.clear();
    this->callSymbols.clear();
    this->returnSymbols.clear();
    set<Symbol> alphabet = this->nondeterministicVPA->getInputAlphabet();
    set<TransitionPDA> transitions = this->nondeterministicVPA->getTransitions();
    for (set<Symbol>::iterator input = alphabet.begin(); input != alphabet.end(); input++) {
        for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
            if (transition->getInput() == *input) {
                if (transition->getPush().size() != 0) {
                    this->callSymbols.insert(*input);
                } else if (transition->getPop().size() != 0) {
                    this->returnSymbols.insert(*input);
                } else {
                    this->internalSymbols.insert(*input);
                }
                break;
            }
        }
    }
}


string VPADeterminizer::buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input) {
    string name = "(" + this->buildStateName(s, r) + ")";
    name += ", " + input.getSymbol();
    return name;
}


void VPADeterminizer::buildStackAlphabet() {
    const set<State>& states = this->nondeterministicVPA->getStates();
    
    // Generate all possible s items
    vector<StatesPair> possibleStatesPairs;
    for (set<State>::iterator state1 = states.begin(); state1 != states.end(); state1++) {
        for (set<State>::iterator state2 = states.begin(); state2 != states.end(); state2++) {
            possibleStatesPairs.push_back(StatesPair(*state1, *state2));
        }
    }
    
    // Generate all possible s components
    vector<SComponent> possibleS;
    int possibleSSize = pow(2, possibleStatesPairs.size());
    for (int i = 0; i < possibleSSize; i++) {
        set<StatesPair> statesPairsSet;
        int mask = 1;
        for (vector<StatesPair>::iterator statesPair = possibleStatesPairs.begin(); statesPair != possibleStatesPairs.end(); statesPair++) {
            if ((i & mask) != 0) {
                statesPairsSet.insert(*statesPair);
            }
            mask <<= 1;
        }
        possibleS.push_back(SComponent(statesPairsSet));
    }
    
    // Generate all possible r components
    vector<RComponent> possibleR;
    int possibleRSize = pow(2, states.size());
    for (int i = 0; i < possibleRSize; i++) {
        set<State> statesSet;
        int mask = 1;
        for (set<State>::iterator state = states.begin(); state != states.end(); state++) {
            if ((i & mask) != 0) {
                statesSet.insert(*state);
            }
            mask <<= 1;
        }
        possibleR.push_back(RComponent(statesSet));
    }
    
    // Generate whole stack alphabet
    set<Symbol> alphabet = this->nondeterministicVPA->getInputAlphabet();
    for (vector<SComponent>::iterator s = possibleS.begin(); s != possibleS.end(); s++) {
        for (vector<RComponent>::iterator r = possibleR.begin(); r != possibleR.end(); r++) {
            this->partialStackAlphabet.push_back(PartialStackData(*s, *r));
            for (set<Symbol>::iterator input = alphabet.begin(); input != alphabet.end(); input++) {
                string stackSymbolName = this->buildStackSymbolName(*s, *r, *input);
                this->deterministicVPA->addStackSymbol(Symbol(stackSymbolName));
            }
        }
    }
}


VPADeterminizer::VPADeterminizer(PDA* automaton) {
    this->nondeterministicVPA = automaton;
}


PDA* VPADeterminizer::determinize() {
    this->deterministicVPA = new PDA();
    DeterminizationUtils::copyInputAlphabet(*this->nondeterministicVPA, *this->deterministicVPA);
    this->divideInputAlphabet();
    this->states.clear();
    this->partialStackAlphabet.clear();
    this->buildStackAlphabet();
    
    const SComponent& initialStateS = this->getSComponentWithStatesIdentity();
    const RComponent& initialStateR(this->nondeterministicVPA->getInitialStates());
    const State& initialState = this->getOrCreateState(initialStateS, initialStateR);
    this->deterministicVPA->addInitialState(initialState);
    
    set<TransitionPDA> transitions = this->nondeterministicVPA->getTransitions();
    
    while (true) {
        StateData* unmarkedStateData = NULL;
        for (map<string, StateData>::iterator stateIter = this->states.begin(); stateIter != this->states.end(); stateIter++) {
            if (!stateIter->second.isMarked) {
                unmarkedStateData = &stateIter->second;
                unmarkedStateData->isMarked = true;
                break;
            }
        }
        if (unmarkedStateData == NULL) {
            break;
        }
        
        // Internal symbols
        for (set<Symbol>::iterator internalSymbol = this->internalSymbols.begin(); internalSymbol != this->internalSymbols.end(); internalSymbol++) {
            SComponent s;
            SComponent& oldS = unmarkedStateData->s;
            for (set<StatesPair>::iterator statesPair = oldS.statesPairs.begin(); statesPair != oldS.statesPairs.end(); statesPair++) {
                for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
                    if (transition->getFrom() == statesPair->state2 && transition->getInput() == *internalSymbol) {
                        s.statesPairs.insert(StatesPair(statesPair->state1, transition->getTo()));
                    }
                }
            }
            RComponent r;
            RComponent& oldR = unmarkedStateData->r;
            for (set<State>::iterator state = oldR.states.begin(); state != oldR.states.end(); state++) {
                for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
                    if (transition->getFrom() == *state && transition->getInput() == *internalSymbol) {
                        r.states.insert(transition->getTo());
                    }
                }
            }
            const State& toState = this->getOrCreateState(s, r);
            const TransitionPDA transition(unmarkedStateData->state, *internalSymbol, toState);
            this->deterministicVPA->addTransition(transition);
        }
        
        // Call symbols
        for (set<Symbol>::iterator callSymbol = this->callSymbols.begin(); callSymbol != this->callSymbols.end(); callSymbol++) {
            RComponent r;
            RComponent& oldR = unmarkedStateData->r;
            for (set<State>::iterator state = oldR.states.begin(); state != oldR.states.end(); state++) {
                for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
                    if (transition->getFrom() == *state && transition->getInput() == *callSymbol) {
                        r.states.insert(transition->getTo());
                    }
                }
            }
            const State& toState = this->getOrCreateState(this->getSComponentWithStatesIdentity(), r);
            string stackSymbolName = this->buildStackSymbolName(unmarkedStateData->s, unmarkedStateData->r, *callSymbol);
            list<Symbol> pop;
            list<Symbol> push(1, Symbol(stackSymbolName));
            const TransitionPDA transition(unmarkedStateData->state, *callSymbol, toState, pop, push);
            this->deterministicVPA->addTransition(transition);
        }
        
        // Return symbols
        for (set<Symbol>::iterator returnSymbol = this->returnSymbols.begin(); returnSymbol != this->returnSymbols.end(); returnSymbol++) {
            SComponent& oldS = unmarkedStateData->s;
            RComponent& oldR = unmarkedStateData->r;
            
            // Empty stack
            list<Symbol> bottomOfStack(1, BOTTOM_OF_STACK_SYMBOL);
            SComponent s;
            for (set<StatesPair>::iterator statesPair = oldS.statesPairs.begin(); statesPair != oldS.statesPairs.end(); statesPair++) {
                for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
                    if (transition->getFrom() == statesPair->state2 && transition->getInput() == *returnSymbol && transition->getPop() == bottomOfStack) {
                        s.statesPairs.insert(StatesPair(statesPair->state1, transition->getTo()));
                    }
                }
            }
            RComponent r;
            for (set<State>::iterator state = oldR.states.begin(); state != oldR.states.end(); state++) {
                for (set<TransitionPDA>::iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
                    if (transition->getFrom() == *state && transition->getInput() == *returnSymbol && transition->getPop() == bottomOfStack) {
                        r.states.insert(transition->getTo());
                    }
                }
            }
            if (r.states.size() != 0 || s.statesPairs.size() != 0) {
                const State& toState = this->getOrCreateState(s, r);
                list<Symbol> pop(1, BOTTOM_OF_STACK_SYMBOL);
                list<Symbol> push;
                const TransitionPDA transition(unmarkedStateData->state, *returnSymbol, toState, pop, push);
                this->deterministicVPA->addTransition(transition);
            }
            
            // Otherwise
            for (set<Symbol>::iterator callSymbol = this->callSymbols.begin(); callSymbol != this->callSymbols.end(); callSymbol++) {
                set<StatesPair> update;
                for (set<StatesPair>::iterator oldSStatesPair = oldS.statesPairs.begin(); oldSStatesPair != oldS.statesPairs.end(); oldSStatesPair++) {
                    if (oldR.states.find(oldSStatesPair->state2) != oldR.states.end()) {
                        const State& state1 = oldSStatesPair->state1;
                        const State& state2 = oldSStatesPair->state2;
                        for (set<TransitionPDA>::iterator transitionPush = transitions.begin(); transitionPush != transitions.end(); transitionPush++) {
                            for (set<TransitionPDA>::iterator transitionPop = transitions.begin(); transitionPop != transitions.end(); transitionPop++) {
                                if (transitionPush->getInput() == *callSymbol && transitionPush->getTo() == state1 &&
                                        transitionPop->getInput() == *returnSymbol && transitionPop->getFrom() == state2 &&
                                        transitionPush->getPush() == transitionPop->getPop()) {
                                    update.insert(StatesPair(transitionPush->getFrom(), transitionPop->getTo()));
                                }
                            }
                        }
                    }
                }
                if (update.size() == 0) {
                    continue;
                }
                
                for (vector<PartialStackData>::iterator partialStackSymbol = this->partialStackAlphabet.begin();
                        partialStackSymbol != this->partialStackAlphabet.end();
                        partialStackSymbol++) {
                    SComponent& stackS = partialStackSymbol->s;
                    RComponent& stackR = partialStackSymbol->r;
                    
                    RComponent r;
                    SComponent s;
                    for (set<StatesPair>::iterator updateItem = update.begin(); updateItem != update.end(); updateItem++) {
                        if (stackR.states.find(updateItem->state1) != stackR.states.end()) {
                            r.states.insert(updateItem->state2);
                        }
                        for (set<StatesPair>::iterator stackSStatesPair = stackS.statesPairs.begin();
                                stackSStatesPair != stackS.statesPairs.end();
                                stackSStatesPair++) {
                            if (stackSStatesPair->state2 == updateItem->state1) {
                                s.statesPairs.insert(StatesPair(stackSStatesPair->state1, updateItem->state2));
                            }
                        }
                    }
                    
                    if (r.states.size() == 0 && s.statesPairs.size() == 0) {
                        continue;
                    }
                    
                    const State& toState = this->getOrCreateState(s, r);
                    string stackSymbolName = this->buildStackSymbolName(stackS, stackR, *callSymbol);
                    list<Symbol> pop(1, Symbol(stackSymbolName));
                    list<Symbol> push;
                    const TransitionPDA transition(unmarkedStateData->state, *returnSymbol, toState, pop, push);
                    this->deterministicVPA->addTransition(transition);
                }
            }
        }
    }
    
    // Final states
    for (map<string, StateData>::iterator stateIter = this->states.begin(); stateIter != this->states.end(); stateIter++) {
        const StateData& stateData = stateIter->second;
        if (DeterminizationUtils::containSomeFinalStateOfAutomaton(stateData.r.states, *this->nondeterministicVPA)) {
            this->deterministicVPA->addFinalState(stateData.state);
        }
    }
    
    return this->deterministicVPA;
}

} /* namespace determinization */
