/*
 * AutomatonFactory.cpp
 *
 *  Created on: Apr 23, 2013
 *      Author: martin
 */

#include "AutomatonFactory.h"

#include <algorithm>

#include "sax/Token.h"
#include "sax/SaxInterface.h"
#include "automaton/AutomatonParser.h"
#include "automaton/exception/AutomatonException.h"

namespace automaton {

using namespace sax;

UnknownAutomaton AutomatonFactory::fromFile(const string& filename) {
	list<Token> tokens;
	SaxInterface::parseFile(filename, tokens);
	return parse(tokens);
}

UnknownAutomaton AutomatonFactory::fromString(const string& str) {
	list<Token> tokens;
	SaxInterface::parseMemory(str, tokens);
	return parse(tokens);
}

UnknownAutomaton AutomatonFactory::parse(list<Token> tokens) {
	AutomatonParser parser;
	return parser.parse(tokens);
}

Automaton* AutomatonFactory::buildAutomaton(const UnknownAutomaton& automaton) {
	if (isFSM(automaton)) {
		return new FSM(buildFSM(automaton));
	} else if (isPDA(automaton)) {
		return new PDA(buildPDA(automaton));
	} else if (isTM(automaton)) {
		return new TM(buildTM(automaton));
	} else {
		throw AutomatonException("Cannot determine automaton type.");
	}
}

void AutomatonFactory::buildCommon(Automaton& automaton,const UnknownAutomaton& unknownAutomaton) {
	set<State>::const_iterator state;
	set<Symbol>::const_iterator symbol;

	const set<State>& states = unknownAutomaton.getStates();
	state = states.begin();
	while (state != states.end()) {
		automaton.addState(*state);
		state++;
	}

	const set<Symbol>& inputAlphabet = unknownAutomaton.getInputAlphabet();
	symbol = inputAlphabet.begin();
	while (symbol != inputAlphabet.end()) {
		automaton.addInputSymbol(*symbol);
		symbol++;
	}

	const set<State>& initialStates = unknownAutomaton.getInitialStates();
	state = initialStates.begin();
	while (state != initialStates.end()) {
		automaton.addInitialState(*state);
		state++;
	}

	const set<State>& finalStates = unknownAutomaton.getFinalStates();
	state = finalStates.begin();
	while (state != finalStates.end()) {
		automaton.addFinalState(*state);
		state++;
	}
}

bool AutomatonFactory::isFSM(const UnknownAutomaton& automaton) {
	if (automaton.getStackAlphabet().size() > 0 || automaton.getStartSymbols().size() > 0
			|| automaton.getTapeAlphabet().size() > 0 || automaton.getBlankSymbol() != Symbol("")) {
		return false;
	}

	const set<UnknownTransition> transitions = automaton.getTransitions();
	set<UnknownTransition>::const_iterator it = transitions.begin();
	while (it != transitions.end()) {
		if (!isFSMTransition(*it)) {
			return false;
		}
		it++;
	}

	return true;
}

bool AutomatonFactory::isFSMTransition(const UnknownTransition& transition) {
	return !(transition.getPop().size() > 0 || transition.getPush().size() > 0 || transition.getOutput() != Symbol("")
			|| transition.getShift() != NOT_SET);
}

FSM AutomatonFactory::buildFSM(const UnknownAutomaton& automaton) {
	FSM fsm;

	buildCommon(fsm, automaton);

	const set<UnknownTransition> transitions = automaton.getTransitions();
	set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		fsm.addTransition(transition->getFrom(), transition->getInput(), transition->getTo());
		transition++;
	}

	return fsm;
}

bool AutomatonFactory::isPDA(const UnknownAutomaton& automaton) {
	if (automaton.getTapeAlphabet().size() > 0 || automaton.getBlankSymbol() != Symbol("")) {
		return false;
	}

	const set<UnknownTransition> transitions = automaton.getTransitions();
	set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		if (!isPDATransition(*transition)) {
			return false;
		}
		transition++;
	}

	return true;
}

bool AutomatonFactory::isPDATransition(const UnknownTransition& transition) {
	return !(transition.getOutput() != Symbol("") || transition.getShift() != NOT_SET);
}

PDA AutomatonFactory::buildPDA(const UnknownAutomaton& automaton) {
	PDA pda;
	buildCommon(pda, automaton);

	const set<Symbol>& stackAlphabet = automaton.getStackAlphabet();
	set<Symbol>::iterator stackSymbol = stackAlphabet.begin();
	while (stackSymbol != stackAlphabet.end()) {
		pda.addStackSymbol(*stackSymbol);
		stackSymbol++;
	}

	pda.setStartSymbols(automaton.getStartSymbols());

	const set<UnknownTransition>& transitions = automaton.getTransitions();

	set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		TransitionPDA tr(transition->getFrom(), transition->getInput(), transition->getTo(), transition->getPop(),
				transition->getPush());
		pda.addTransition(tr);
		transition++;
	}

	return pda;
}

bool AutomatonFactory::isTM(const UnknownAutomaton& automaton) {
	if (automaton.getStackAlphabet().size() > 0 || automaton.getStartSymbols().size() > 0) {
		return false;
	}

	const set<UnknownTransition>& transitions = automaton.getTransitions();

	set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		if (!isTMTransition(*transition)) {
			return false;
		}
		transition++;
	}

	return true;

}

bool AutomatonFactory::isTMTransition(const UnknownTransition& transition) {
	return !(transition.getPop().size() > 0 || transition.getPush().size() > 0);

}

TM AutomatonFactory::buildTM(const UnknownAutomaton& automaton) {
	TM tm;

	const set<Symbol>& tapeAlphabet = automaton.getTapeAlphabet();

	set<Symbol>::const_iterator tapeSymbol = tapeAlphabet.begin();
	while (tapeSymbol != tapeAlphabet.end()) {
		tm.addTapeSymbol(*tapeSymbol);
		tapeSymbol++;
	}

	buildCommon(tm, automaton);

	tm.setBlankSymbol(automaton.getBlankSymbol());

	const set<UnknownTransition>& transitions = automaton.getTransitions();

	set<UnknownTransition>::const_iterator transition = transitions.begin();
	while (transition != transitions.end()) {
		TransitionTM tr(transition->getFrom(), transition->getInput(), transition->getTo(), transition->getOutput(),
				transition->getShift());
		tm.addTransition(tr);
		transition++;
	}

	return tm;

}

} /* namespace automaton */
