/*
 * AutomatonParser.cpp
 *
 *  Created on: Oct 12, 2013
 *      Author: Martin Zak
 */

#include "AutomatonParser.h"

#include "../sax/ParserException.h"
#include "Shift.h"

namespace automaton {

using namespace std;
using namespace sax;

UnknownAutomaton AutomatonParser::parse(list<Token> &input) {
	UnknownAutomaton automaton;

	popToken(input, Token::START_ELEMENT, "automaton");

	while (true) {
		if (isToken(input, Token::END_ELEMENT, "automaton")) {
			break;
		} else if (isToken(input, Token::START_ELEMENT, "states")) {
			parseStates(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "inputAlphabet")) {
			parseInputAlphabet(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "initialStates")) {
			parseInitialStates(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "finalStates")) {
			parseFinalStates(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "stackAlphabet")) {
			parseStackAlphabet(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "startSymbols")) {
			parseStartSymbols(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "tapeAlphabet")) {
			parseTapeAlphabet(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "blankSymbol")) {
			parseBlankSymbol(input, automaton);
		} else if (isToken(input, Token::START_ELEMENT, "transitions")) {
			parseTransitions(input, automaton);

		} else {
			throw ParserException(Token("automaton", Token::END_ELEMENT), input.front());
		}
	}

	popToken(input, Token::END_ELEMENT, "automaton");
	return automaton;
}

void AutomatonParser::parseStates(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "states");
	while (isToken(input, Token::START_ELEMENT, "state")) {
		automaton.addState(parseState(input));
	}
	popToken(input, Token::END_ELEMENT, "states");
}

void AutomatonParser::parseInputAlphabet(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "inputAlphabet");
	while (isToken(input, Token::START_ELEMENT, "symbol")) {
		automaton.addInputSymbol(parseSymbol(input));
	}
	popToken(input, Token::END_ELEMENT, "inputAlphabet");
}

void AutomatonParser::parseInitialStates(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "initialStates");
	while (isToken(input, Token::START_ELEMENT, "state")) {
		automaton.addInitialState(parseState(input));
	}
	popToken(input, Token::END_ELEMENT, "initialStates");
}

void AutomatonParser::parseFinalStates(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "finalStates");
	while (isToken(input, Token::START_ELEMENT, "state")) {
		automaton.addFinalState(parseState(input));
	}
	popToken(input, Token::END_ELEMENT, "finalStates");
}

void AutomatonParser::parseStackAlphabet(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "stackAlphabet");
	while (isToken(input, Token::START_ELEMENT, "symbol")) {
		automaton.addStackSymbol(parseSymbol(input));
	}
	popToken(input, Token::END_ELEMENT, "stackAlphabet");
}

void AutomatonParser::parseStartSymbols(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "startSymbols");
	list<Symbol> startSymbols;
	while (isToken(input, Token::START_ELEMENT, "symbol")) {
		startSymbols.push_back(parseSymbol(input));
	}
	automaton.setStartSymbols(startSymbols);
	popToken(input, Token::END_ELEMENT, "startSymbols");
}

void AutomatonParser::parseTapeAlphabet(list<Token>& input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "tapeAlphabet");
	while (isToken(input, Token::START_ELEMENT, "symbol")) {
		automaton.addTapeSymbol(parseSymbol(input));
	}
	popToken(input, Token::END_ELEMENT, "tapeAlphabet");
}

void AutomatonParser::parseBlankSymbol(list<Token>& input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "blankSymbol");
	if (input.front().getType() == Token::CHARACTER) {
		automaton.setBlankSymbol(Symbol(input.front().getData()));
		input.pop_front();
	} else {
		throw ParserException(Token("", Token::CHARACTER), input.front());
	}
	popToken(input, Token::END_ELEMENT, "blankSymbol");
}

void AutomatonParser::parseTransitions(list<Token> &input, UnknownAutomaton& automaton) {
	popToken(input, Token::START_ELEMENT, "transitions");
	while (isToken(input, Token::START_ELEMENT, "transition")) {
		automaton.addTransition(parseTransition(input));
	}
	popToken(input, Token::END_ELEMENT, "transitions");
}

void AutomatonParser::parsePop(list<Token>& input, UnknownTransition* transition) {
	popToken(input, Token::START_ELEMENT, "pop");
	while (isToken(input, Token::START_ELEMENT, "symbol")) {
		transition->addPop(parseSymbol(input));
	}
	popToken(input, Token::END_ELEMENT, "pop");
}

void AutomatonParser::parsePush(list<Token>& input, UnknownTransition* transition) {
	popToken(input, Token::START_ELEMENT, "push");
	while (isToken(input, Token::START_ELEMENT, "symbol")) {
		transition->addPush(parseSymbol(input));
	}
	popToken(input, Token::END_ELEMENT, "push");
}

State AutomatonParser::parseState(list<Token> &input, string tagName) {
	popToken(input, Token::START_ELEMENT, tagName);

	if (input.front().getType() == Token::CHARACTER) {
		State state(input.front().getData());
		input.pop_front();
		popToken(input, Token::END_ELEMENT, tagName);
		return state;
	} else {
		throw ParserException(Token("", Token::CHARACTER), input.front());
	}
}

Symbol AutomatonParser::parseSymbol(list<Token>& input, string tagName) {
	popToken(input, Token::START_ELEMENT, tagName);

	if (input.front().getType() == Token::CHARACTER) {
		Symbol symbol(input.front().getData());
		input.pop_front();
		popToken(input, Token::END_ELEMENT, tagName);
		return symbol;
	} else if(isToken(input,Token::END_ELEMENT, tagName)){
		input.pop_front();
		return Symbol("");
	} else if(isToken(input,Token::START_ELEMENT, "eps")) {
		input.pop_front();
		popToken(input, Token::END_ELEMENT,"eps");
		popToken(input, Token::END_ELEMENT, tagName);
		return Symbol("");
	} else {
		throw ParserException(Token("", Token::CHARACTER), input.front());
	}
}

Shift AutomatonParser::parseShift(list<Token>& input) {
	Shift shift;

	popToken(input, Token::START_ELEMENT, "shift");
	if (input.front().getType() == Token::CHARACTER) {
		if (input.front().getData() == "left")
			shift = LEFT;
		if (input.front().getData() == "right")
			shift = RIGHT;
		if (input.front().getData() == "none")
			shift = NONE;
		input.pop_front();
	} else {
		throw sax::ParserException(sax::Token("", sax::Token::CHARACTER), input.front());
	}
	popToken(input, Token::END_ELEMENT, "shift");

	return shift;

}

UnknownTransition AutomatonParser::parseTransition(list<Token>& input) {
	UnknownTransition transition;

	popToken(input, Token::START_ELEMENT, "transition");
	while (true) {
		if (isToken(input, Token::END_ELEMENT, "transition")) {
			break;
		} else if (isToken(input, Token::START_ELEMENT, "from")) {
			transition.setFrom(parseState(input, "from"));
		} else if (isToken(input, Token::START_ELEMENT, "input")) {
			transition.setInput(parseSymbol(input, "input"));
		} else if (isToken(input, Token::START_ELEMENT, "to")) {
			transition.setTo(parseState(input, "to"));
		} else if (isToken(input, Token::START_ELEMENT, "pop")) {
			parsePop(input, &transition);
		} else if (isToken(input, Token::START_ELEMENT, "push")) {
			parsePush(input, &transition);
		} else if (isToken(input, Token::START_ELEMENT, "output")) {
			transition.setOutput(parseSymbol(input, "output"));
		} else if (isToken(input, Token::START_ELEMENT, "shift")) {
			transition.setShift(parseShift(input));
		} else {
			throw ParserException(Token("transitions", Token::END_ELEMENT), input.front());
		}
	}
	popToken(input, Token::END_ELEMENT, "transition");

	return transition;
}

bool AutomatonParser::isToken(list<Token>& input, Token::TokenType type, string data) {
	return input.front().getType() == type && input.front().getData() == data;
}

void AutomatonParser::popToken(list<Token>& input, Token::TokenType type, string data) {
	if (isToken(input, type, data)) {
		input.pop_front();
	} else {
		throw ParserException(Token(data, type), input.front());
	}
}

} /* namespace automaton */
