/*
 * TM.cpp
 *
 *  Created on: Apr 24, 2013
 *      Author: martin
 */

#include "TM.h"
#include "../AutomatonPrinter.h"
#include "../exception/AutomatonException.h"

namespace automaton {

using namespace std;
using namespace alphabet;

TM::TM() :
		blankSymbol(Symbol("")) {
}

void TM::addInputSymbol(const Symbol& symbol) {
	if (tapeAlphabet.find(symbol) == tapeAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + symbol.getSymbol() + "\" is not in tape alphabet.");
	}

	if (symbol == blankSymbol) {
		throw AutomatonException("Input symbol \"" + symbol.getSymbol() + "\" cannot be blank symbol.");
	}

	std::pair<std::set<Symbol>::iterator, bool> ret = inputAlphabet.insert(symbol);
	if (!ret.second)
		throw AutomatonException("Input symbol \"" + symbol.getSymbol() + "\" already exists.");
}

void TM::removeState(const State& state) {
	for (set<State>::const_iterator initialState = initialStates.begin(); initialState != initialStates.end();
			initialState++) {
		if (*initialState == state) {
			throw AutomatonException("State \"" + state.getName() + "\" is initial state.");
		}
	}

	for (set<State>::const_iterator finalState = finalStates.begin(); finalState != finalStates.end(); finalState++) {
		if (*finalState == state) {
			throw AutomatonException("State \"" + state.getName() + "\" is final state.");
		}
	}

	for (set<TransitionTM>::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (state == t->getFrom() || state == t->getTo())
			throw AutomatonException("State \"" + state.getName() + "\" is used.");
	}

	int removed = states.erase(state);
	if (!removed)
		throw AutomatonException("State \"" + state.getName() + "\" doesn't exist.");
}

void TM::removeInputSymbol(const Symbol& symbol) {
	int removed = inputAlphabet.erase(symbol);
	if (!removed)
		throw AutomatonException("Input symbol \"" + symbol.getSymbol() + "\" doesn't exist.");
}

void TM::addTapeSymbol(const Symbol& symbol) {
	pair<set<Symbol>::iterator, bool> ret = tapeAlphabet.insert(symbol);
	if (!ret.second)
		throw AutomatonException("Tape symbol \"" + symbol.getSymbol() + "\" already exists.");
}

void TM::removeTapeSymbol(const Symbol& symbol) {
	for (set<Symbol>::const_iterator it = inputAlphabet.begin(); it != inputAlphabet.end(); it++) {
		if (symbol == (*it)) {
			throw AutomatonException("Tape symbol \"" + symbol.getSymbol() + "\" is in input alphabet.");
		}
	}

	for (set<TransitionTM>::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (symbol == t->getInput() || symbol == t->getOutput())
			throw AutomatonException("Tape symbol \"" + symbol.getSymbol() + "\" is used in transition.");
	}

	int removed = inputAlphabet.erase(symbol);
	if (!removed)
		throw AutomatonException("Tape symbol \"" + symbol.getSymbol() + "\" doesn't exist.");

}

const std::set<Symbol>& TM::getTapeAlphabet() const {
	return tapeAlphabet;
}

void TM::addTransition(const TransitionTM& transition) {
	if (states.find(transition.getFrom()) == states.end()) {
		throw AutomatonException("State \"" + transition.getFrom().getName() + "\" doesn't exist.");
	}

	if (transition.getInput().getSymbol() != "") {
		if (tapeAlphabet.find(transition.getInput()) == tapeAlphabet.end()) {
			throw AutomatonException("Tape symbol \"" + transition.getInput().getSymbol() + "\" doesn't exist.");
		}
	}

	if (states.find(transition.getTo()) == states.end()) {
		throw AutomatonException("State \"" + transition.getTo().getName() + "\" doesn't exist.");
	}

	if (tapeAlphabet.find(transition.getOutput()) == tapeAlphabet.end()) {
		throw AutomatonException("Tape symbol  \"" + transition.getOutput().getSymbol() + "\" doesn't exist.");
	}

	pair<set<TransitionTM>::iterator, bool> ret = transitions.insert(transition);
	if (!ret.second) {
		throw AutomatonException("Transition already exists.");
	}

}

void TM::removeTransition(const TransitionTM& transition) {
	int removed = transitions.erase(transition);
	if (!removed) {
		throw AutomatonException("Transition doesn't exist.");
	}
}

const std::set<TransitionTM>& TM::getTransitions() const {
	return transitions;
}

void TM::setBlankSymbol(const Symbol& symbol) {
	for (set<Symbol>::const_iterator it = inputAlphabet.begin(); it != inputAlphabet.end(); it++) {
		if (symbol == (*it)) {
			throw AutomatonException("Blank symbol \"" + symbol.getSymbol() + "\" is in input alphabet.");
		}
	}

	if (tapeAlphabet.find(symbol) == tapeAlphabet.end())
		throw AutomatonException("Blank symbol \"" + symbol.getSymbol() + "\" is not in tape alphabet.");

	blankSymbol = symbol;

}
const Symbol& TM::getBlankSymbol() const {
	return blankSymbol;
}

void TM::toXML(std::ostream& out) const {
	AutomatonPrinter::toXML(*this, out);
}

} /* namespace automaton */
