/*
 * Grammar.h
 *
 *  Created on: Nov 2, 2013
 *      Author: Martin Zak
 */

#ifndef GRAMMAR_H_
#define GRAMMAR_H_

#include <set>
#include <list>

#include "../alphabet/Symbol.h"
#include "Rule.h"

namespace grammar {

using namespace std;
using namespace alphabet;

/**
 * Abstract class representing grammar. Implements methods for manipulation
 * with terminal and nonterminal symbols. Only method, which is not implemented
 * is grammar specific isValidRule.
 */
class Grammar {
protected:
	set<Symbol> terminalSymbols;
	set<Symbol> nonTerminalSymbols;
	set<Rule> rules;
	Symbol startSymbol;

	/**
	 * Checks that rule has valid form for this grammar.
	 * @param rule Rule to check
	 * @return true when rule is valid, false when not
	 */
	virtual bool isValidRule(const Rule& rule) const = 0;
public:
	Grammar();
	Grammar(const set<Symbol>& nonTerminalSymbols, const set<Symbol>& terminalSymbols, const Symbol& startSymbol);
	virtual ~Grammar();

	/**
	 * Adds terminal symbol to the grammar.
	 * @param symbol Symbol to add
	 * @throws GrammarException when symbol is already terminal or nonterminal symbol
	 */
	void addTerminalSymbol(const Symbol& symbol);

	/**
	 * Removes terminal symbol from the grammar.
	 * @param symbol Symbol to remove
	 * @throws GrammarException when symbol is used in rule or isn't terminal symbol
	 */
	void removeTerminalSymbol(const Symbol& symbol);

	/**
	 * @return set of terminal symbols
	 */
	const set<Symbol>& getTerminalSymbols() const;

	/**
	 * Adds nonterminal symbol to the grammar.
	 * @param symbol Symbol to add
	 * @throws GrammarException when symbol is already terminal or nonterminal symbol
	 */
	void addNonTerminalSymbol(const Symbol& symbol);

	/**
	 * Removes nonterminal symbol from the grammar.
	 * @param symbol Symbol to remove
	 * @throws GrammarException when symbol is used in rule or isn't nonterminal symbol
	 */
	void removeNonTerminalSymbol(const Symbol& symbol);

	/**
	 * @return set of nonterminal symbols
	 */
	const set<Symbol>& getNonTerminalSymbols() const;

	/**
	 * Adds Rule to the grammar. Perform checks that rule is valid for this grammar.
	 * @param rule Rule to add
	 * @throws GrammarException when Rule is not valid for this grammar or when rule already exists
	 */
	void addRule(const Rule& rule);

	/**
	 * Removes Rule from the grammar.
	 * @param rule Rule to remove
	 * @throws GrammarException when Rule doesn't exist
	 */
	void removeRule(const Rule& rule);

	/**
	 * @return set of Rules
	 */
	const set<Rule>& getRules() const;

	/**
	 * @return start symbol
	 */
	const Symbol& getStartSymbol() const;

	/**
	 * Sets the start symbol. Checks that symbol is nonterminal symbol.
	 * @param symbol nonterminal Symbol
	 * @throws GrammarException when symbol isn't nonterminal symbol
	 */
	void setStartSymbol(const Symbol& symbol);

	/**
	 * Prints the XML representation of grammar to the given output stream.
	 * @param out output stream to print to
	 */
	void toXML(ostream& out) const;
};

} /* namespace grammar */
#endif /* GRAMMAR_H_ */
