/*
 * RegExpParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpParser.h"
#include "../sax/ParserException.h"

namespace regexp {

RegExp RegExpParser::parse(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "regexp");
	RegExp regexp;
	regexp.setRegExp(parseElement(input));
	popToken(input, Token::END_ELEMENT, "regexp");

	return regexp;
}

RegExpElement* RegExpParser::parseElement(list<Token>& input) {
	if (isToken(input, Token::START_ELEMENT, "symbol")) {
		return parseSymbol(input);
	} else if (isToken(input, Token::START_ELEMENT, "iteration")) {
		return parseIteration(input);
	} else if (isToken(input, Token::START_ELEMENT, "alternation")) {
		return parseAlternation(input);
	} else if (isToken(input, Token::START_ELEMENT, "concatenation")) {
		return parseConcatenation(input);
	} else {
		return NULL;
	}
}

Alternation* RegExpParser::parseAlternation(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "alternation");

	Alternation* alternation = new Alternation;
	parseContent(input, alternation->getElements());

	popToken(input, Token::END_ELEMENT, "alternation");
	return alternation;
}

Concatenation* RegExpParser::parseConcatenation(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "concatenation");

	Concatenation* concatenation = new Concatenation();
	parseContent(input, concatenation->getElements());

	popToken(input, Token::END_ELEMENT, "concatenation");
	return concatenation;

}

Iteration* RegExpParser::parseIteration(list<Token>& input) {
	popToken(input, Token::START_ELEMENT, "iteration");

	Iteration* iteration = new Iteration();
	iteration->setElement(parseElement(input));

	popToken(input, Token::END_ELEMENT, "iteration");
	return iteration;
}

void RegExpParser::parseContent(list<Token>& input, list<RegExpElement*>& elements) {
	while (true) {
		if (isToken(input, Token::START_ELEMENT, "symbol")) {
			elements.push_back(parseSymbol(input));
		} else if (isToken(input, Token::START_ELEMENT, "iteration")) {
			elements.push_back(parseIteration(input));
		} else if (isToken(input, Token::START_ELEMENT, "alternation")) {
			elements.push_back(parseAlternation(input));
		} else if (isToken(input, Token::START_ELEMENT, "concatenation")) {
			elements.push_back(parseConcatenation(input));
		} else {
			return;
		}
	}

}

RegExpSymbol* RegExpParser::parseSymbol(list<Token>& input, string tagName) {
	popToken(input, Token::START_ELEMENT, tagName);

	if (input.front().getType() == Token::CHARACTER) {
		RegExpSymbol* symbol = new RegExpSymbol(input.front().getData());
		input.pop_front();
		popToken(input, Token::END_ELEMENT, tagName);
		return symbol;
	} else if (isToken(input, Token::END_ELEMENT, tagName)) {
		input.pop_front();
		return new RegExpSymbol("");
	} else if (isToken(input, Token::START_ELEMENT, "eps")) {
		input.pop_front();
		popToken(input, Token::END_ELEMENT, "eps");
		popToken(input, Token::END_ELEMENT, tagName);
		return new RegExpSymbol("");
	} else {
		throw ParserException(Token("", Token::CHARACTER), input.front());
	}
}

bool RegExpParser::isToken(list<Token>& input, Token::TokenType type, string data) {
	return input.front().getType() == type && input.front().getData() == data;
}

void RegExpParser::popToken(list<Token>& input, Token::TokenType type, string data) {
	if (isToken(input, type, data)) {
		input.pop_front();
	} else {
		throw ParserException(Token(data, type), input.front());
	}
}

} /* namespace regexp */

